"""
==========
LAISSEZ
==========
"""


from typing import Callable, Dict, List, Optional
from eth_account import Account
from httpx import AsyncClient, Response
from pydantic import BaseModel
from x402.clients.base import MissingRequestConfigError, PaymentError, decode_x_payment_response, x402Client
from x402.clients.httpx import HttpxHooks, x402HttpxClient as BaseX402HttpxClient
from x402.types import x402PaymentRequiredResponse
import os
import csv
from datetime import datetime, timezone
from laissez.common import check_and_create_wallet


class PaymentLog(BaseModel):
    transaction_hash: str
    network: str
    description: str
    paid_to: str
    paid_by: str
    amount: str
    unit: str




class LaissezClient(BaseX402HttpxClient):
    def __init__(self, wallet: Optional[Account] = None, **kwargs):

        wallet = check_and_create_wallet(wallet)

        class LaissezHttpxHooks(HttpxHooks):
            def __init__(self, client: x402Client):
                super().__init__(client=client)

            async def on_response(self, response: Response) -> Response:
                if response.status_code != 402:
                    return response

                if response.request and response.request.extensions.get("x402_retry"):
                    return response

                try:
                    if not response.request:
                        raise MissingRequestConfigError("Missing request configuration")

                    await response.aread()
                    data = response.json()
                    payment_response = x402PaymentRequiredResponse(**data)

                    selected_requirements = self.client.select_payment_requirements(
                        payment_response.accepts
                    )

                    payment_header = self.client.create_payment_header(
                        selected_requirements, payment_response.x402_version
                    )

                    request = response.request
                    request.headers["X-Payment"] = payment_header
                    request.headers["Access-Control-Expose-Headers"] = "X-Payment-Response"
                    request.extensions["x402_retry"] = True

                    async with AsyncClient(follow_redirects=True, timeout=30.0) as client:
                        retry_response = await client.send(request)

                    if 200 <= retry_response.status_code < 300:
                        payment_info = "with unknown details"
                        decoded_header = {}
                        
                        if 'x-payment-response' in retry_response.headers:
                            try:
                                decoded_header = decode_x_payment_response(retry_response.headers['x-payment-response'])
                                payment_info = f"Transaction: {decoded_header.get('transaction')}\n\tPayer: {decoded_header.get('payer')}\n\tNetwork: {decoded_header.get('network')}"
                            except Exception as e:
                                payment_info = f"with settlement details that could not be decoded: {e}"

                        print(f"[Client] Payment successful:\n\t{payment_info}")

                        # post to laissez db
                        payment_log = PaymentLog(
                            transaction_hash=decoded_header.get("transaction", "unknown"),
                            network=selected_requirements.network,
                            description=selected_requirements.description,
                            paid_to=selected_requirements.pay_to,
                            paid_by=decoded_header.get("payer", self.client.account.address),
                            amount=selected_requirements.max_amount_required,
                            unit=selected_requirements.extra.get("name", "unknown"),
                        )

                        # Ensure directory exists
                        cwd = os.getcwd()
                        laissez_dir = os.path.join(cwd, ".laissez")
                        filepath = os.path.join(laissez_dir, "payments.csv")
                        os.makedirs(laissez_dir, exist_ok=True)

                        write_header = not os.path.exists(filepath)

                        # Prepare the row
                        row = {
                            "timestamp": datetime.now(timezone.utc).isoformat(),
                            "transaction_hash": payment_log.transaction_hash,
                            "network": payment_log.network,
                            "description": payment_log.description,
                            "paid_to": payment_log.paid_to,
                            "paid_by": payment_log.paid_by,
                            "amount": payment_log.amount,
                            "unit": payment_log.unit,
                        }

                        # Write to CSV
                        with open(filepath, mode="a", newline="", encoding="utf-8") as csvfile:
                            writer = csv.DictWriter(csvfile, fieldnames=list(row.keys()))
                            if write_header:
                                writer.writeheader()
                            writer.writerow(row)

                        response.status_code = retry_response.status_code
                        response.headers = retry_response.headers
                        response._content = retry_response.content
                        response.request = retry_response.request
                    
                        return response
                    else:
                        await retry_response.aread()
                        message = f"Payment failed with status code {retry_response.status_code}"
                        try:
                            data = retry_response.json()
                            if "error" in data:
                                message += f": {data['error']}"
                        except Exception:
                            message += f". Body: {retry_response.text}"
                        raise PaymentError(message)

                except PaymentError as e:
                    raise e
                except Exception as e:
                    raise PaymentError(f"Failed to handle payment: {str(e)}") from e


        def laissez_hooks(
            account: Account
        ) -> Dict[str, List[Callable]]:
            client = x402Client(account=account)
            hooks = LaissezHttpxHooks(client=client)
            return {
                "request": [hooks.on_request],
                "response": [hooks.on_response],
            }


        super().__init__(account=wallet, follow_redirects=True, timeout=30.0, **kwargs)
        self.event_hooks = laissez_hooks(account=wallet)


"""
==========
LAISSEZ
==========
"""

import asyncio
import base64
import json
from fastmcp import FastMCP
from fastmcp.server.http import StarletteWithLifespan
from eth_account import Account
from pydantic import BaseModel, Field
from typing import List, Literal, Optional, cast
from starlette.responses import JSONResponse
from starlette.types import Message, Receive, Scope, Send
from x402.facilitator import FacilitatorClient
from x402.common import find_matching_payment_requirements, process_price_to_atomic_amount, x402_VERSION
from x402.types import PaymentPayload, PaymentRequirements, x402PaymentRequiredResponse, SupportedNetworks
from laissez.common import check_and_create_wallet


class PaidTool(BaseModel):
    """
    Configure a paid tool for an MCP server.

    Args:
        name: The name of the tool
        price: The price of the tool in USDC. Defaults to 0.001.
        network: The network to use for the tool. Defaults to 'base-sepolia'.
        description: The description of the tool
    """
    name: str = Field(..., description="The name of the tool")
    price: float = Field(0.001, description="The price of the tool in USDC")
    network: Optional[Literal['base-sepolia', 'base']] = Field('base-sepolia', description="The network to use for the tool")
    description: str = Field(..., description="The description of the tool")



class RequestReplay:
    def __init__(self, receive: Receive, body: bytes):
        self._receive = receive
        self._body = body
        self._called = False

    async def __call__(self) -> Message:
        if not self._called:
            self._called = True
            return {"type": "http.request", "body": self._body, "more_body": False}
        return await self._receive()


class LaissezMcpServerMiddleware:
    def __init__(self, app: StarletteWithLifespan, tools: List[PaidTool], wallet: Account):
        self.app = app
        self.tools = tools
        self.wallet = wallet
        self.facilitator = FacilitatorClient()
        self.settlement_lock = asyncio.Lock()

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return
        
        method = scope.get("method", "")
        path = scope.get("path", "")
        scheme = scope.get("scheme", "https")

        # Check if it's an MCP POST request
        if method != "POST" or not path.startswith("/mcp"):
            print(f"[x402] Not an MCP request, passing through")
            await self.app(scope, receive, send)
            return
        
        # Parse the MCP request body
        body = b''
        more_body = True
        while more_body:
            message = await receive()
            if message['type'] == 'http.request':
                body += message.get('body', b'')
                more_body = message.get('more_body', False)
        
        receive_replay = RequestReplay(receive, body)

        # Parse the tool action
        tool_action = None
        if body:
            try:
                mcp_request = json.loads(body)
                tool_action = mcp_request.get("method")
            except json.JSONDecodeError:
                print(f"[x402] Failed to parse MCP request: {body}")

        if tool_action != "tools/call":
            print(f"[x402] Not a tool call, passing through")
            await self.app(scope, receive_replay, send)
            return

        # Parse the tool name
        tool_parameters = mcp_request.get("params", {})
        tool_name = tool_parameters.get("name")

        if not tool_name:
            print(f"[x402] No tool name found, passing through")
            await self.app(scope, receive_replay, send)
            return

        print(f"[x402] Found tool name: {tool_name}")

        # Check if the tool is in the payment config
        tool_names = [tool.name for tool in self.tools]
        if tool_name not in tool_names:
            print(f"[x402] Tool {tool_name} not in payment config, passing through")
            await self.app(scope, receive_replay, send)
            return

        # Build the payment requirements
        headers = { k.decode().lower(): v.decode() for k, v in scope.get("headers", []) }
        tool = self.tools[tool_names.index(tool_name)]
        price = str(tool.price)
        network = tool.network

        try:
            max_amount, asset_address, eip712 = process_price_to_atomic_amount(price, network)
        except ValueError as e:
            error_response = JSONResponse(content={"error": f"Invalid price configuration: {str(e)}"}, status_code=500)
            await error_response(scope, receive_replay, send)
            return

        host = headers.get("host", "")
        resource_url = f"{scheme}://{host}{path}"
        
        payment_requirements = [
            PaymentRequirements(
                scheme="exact",
                network=cast(SupportedNetworks, network),
                asset=asset_address,
                max_amount_required=max_amount,
                resource=resource_url,
                description=tool.description,
                mime_type="application/json",
                pay_to=self.wallet.address,
                max_timeout_seconds=60,
                extra=eip712,
            )
        ]

        # Check if payment is required
        if 'x-payment' not in headers:
            print(f"[x402] Payment required for tool: {tool_name}")

            response_data = x402PaymentRequiredResponse(
                x402_version=x402_VERSION,
                accepts=payment_requirements,
                error=f"Payment required for {tool_name}",
            ).model_dump(by_alias=True)

            error_response = JSONResponse(content=response_data, status_code=402)
            await error_response(scope, receive_replay, send)
            return

        # Decode the X-Payment header
        try:
            payment_dictionary = json.loads(base64.b64decode(headers['x-payment']))
            payment = PaymentPayload(**payment_dictionary)
        except Exception as e:
            print(f"[x402] Failed to decode X-Payment header: {e}")
            error_response = JSONResponse(content={"error": f"Invalid payment header format: {str(e)}"}, status_code=400)
            await error_response(scope, receive_replay, send)
            return
            
        selected_payment_requirements = find_matching_payment_requirements(
            payment_requirements=payment_requirements,
            payment=payment
        )


        if not selected_payment_requirements:
            print(f"[x402] No matching payment requirements found, passing through")
            error_response = JSONResponse(content={"error": f"No matching payment requirements found"}, status_code=400)
            await error_response(scope, receive_replay, send)
            return


        verify_response = await self.facilitator.verify(payment, selected_payment_requirements)

        if not verify_response.is_valid:
            print(f"[x402] Payment verification failed: {verify_response.invalid_reason}")
            error_response = JSONResponse(content={"error": f"Payment verification failed: {verify_response.invalid_reason}"}, status_code=402)
            await error_response(scope, receive_replay, send)
            return


        async def send_wrapper(message: Message):
            if message['type'] == 'http.response.start':
                status_code = message['status']
                if 200 <= status_code < 300:
                    try:
                        settle_response = None
                        async with self.settlement_lock:
                            settle_response = await self.facilitator.settle(payment, selected_payment_requirements)

                        if settle_response and settle_response.success:
                            settlement_header = base64.b64encode(
                                settle_response.model_dump_json(by_alias=True).encode('utf-8')
                            ).decode('utf-8')
                            message['headers'].append(
                                (b'X-Payment-Response', settlement_header.encode('utf-8'))
                            )
                            message['headers'].append(
                                (b'Access-Control-Expose-Headers', b'X-Payment-Response')
                            )
                        else:
                            print(f"[x402] Settlement failed: {settle_response.error_reason if settle_response else 'Unknown reason'}")
                            error_body = json.dumps({"error": "Payment settlement failed"}).encode('utf-8')
                            await send({
                                'type': 'http.response.start',
                                'status': 402,
                                'headers': [
                                    (b'content-type', b'application/json'),
                                    (b'content-length', str(len(error_body)).encode('utf-8'))
                                ]
                            })
                            await send({
                                'type': 'http.response.body',
                                'body': error_body,
                                'more_body': False
                            })
                            return
                    except Exception as e:
                        print(f"[x402] Settlement error: {e}")
                        error_body = json.dumps({"error": "Internal server error during payment settlement"}).encode('utf-8')
                        await send({
                            'type': 'http.response.start',
                            'status': 500,
                            'headers': [
                                (b'content-type', b'application/json'),
                                (b'content-length', str(len(error_body)).encode('utf-8'))
                            ]
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': error_body,
                            'more_body': False
                        })
                        return
            await send(message)

        print(f"[x402] Payment verified, passing to application")
        
        await self.app(scope, receive_replay, send_wrapper)





async def create_paid_mcp_app(mcp: FastMCP, tools: List[PaidTool], wallet: Optional[Account] = None) -> StarletteWithLifespan:

    wallet = check_and_create_wallet(wallet)
    mcp_tools = await mcp.get_tools()
    processed_tools = []
    for tool in tools:
        if tool.name not in mcp_tools:
            print(f"[Laissez] Tool {tool.name} not found in MCP server. Skipping...")
            continue
        else:
            processed_tools.append(tool)
        

    app = mcp.http_app(
        transport='streamable-http',
        stateless_http=True
    )
    app.add_middleware(LaissezMcpServerMiddleware, tools=processed_tools, wallet=wallet)
    return app


"""
==========
LAISSEZ
==========

EXAMPLE: MCP Client with Laissez

This example shows how to use Laissez with the MCP client from the Pydantic AI SDK.

Pre-requisites:
* Ensure that you are running the paid MCP server from `examples/mcp/paid-mcp-server.py`.
    * You can do so from the root of the project by running `uv run examples/mcp/paid-mcp-server.py`.
* Update `BASE_URL` in this file to point to the URL of this example paid MCP server
* Set `WALLET_PRIVATE_KEY` in `.env` equal to the private key for your EVM-compatible account
* Ensure this account has sufficient funds to pay for the MCP calls

To run:
* Run this file from the root of the project by running `uv run examples/mcp/mcp-client.py`.
"""

import asyncio
from pydantic_ai import Agent
from pydantic_ai.mcp import MCPServerStreamableHTTP
from laissez.client import LaissezClient
from dotenv import load_dotenv

load_dotenv()

BASE_URL = 'http://localhost:8000'
laissez_client = LaissezClient()


async def main():

    server = MCPServerStreamableHTTP(
        url=f'{BASE_URL}/mcp',
        http_client=laissez_client   
    )

    agent = Agent(
        model='openai:gpt-5-mini',
        instructions='Show your working out',
        toolsets=[server]
    )

    result = await agent.run("Roll a dice twice then multiply the two results and return the final result")
    print(result.output)


if __name__ == '__main__':
    asyncio.run(main())


import os
import secrets
from eth_account import Account
from typing import Optional


def check_and_create_wallet(wallet: Optional[Account] = None) -> Account:
    if wallet is None:
        wallet_pk = os.getenv("WALLET_PRIVATE_KEY")
        if wallet_pk:
            wallet = Account.from_key(wallet_pk)
            print(f"[Laissez] Loaded wallet from .env file")
            print(f"[Laissez] Wallet address: {wallet.address}")
        else:
            wallet_creation_prompt = """
No wallet found. 
If you have a wallet private key, please set WALLET_PRIVATE_KEY in the .env file and run the server again.
If you don't have a wallet private key, you can create one now.
Note: If you choose to create a wallet, Laissez will add the wallet private key to the .env file.

Would you like to create a wallet now? (Y/n) >>> """
            wallet_creation_response = input(wallet_creation_prompt)
            if wallet_creation_response.lower() == "y" or wallet_creation_response.lower() == "":
                random_string = secrets.token_urlsafe(32)
                wallet = Account.create(random_string)
                env_file = ".env"
                with open(env_file, "a", encoding="utf-8") as f:
                    f.write(f"\nWALLET_PRIVATE_KEY={wallet.key.hex()}\n")
                    print("[Laissez] Wallet created and saved to .env file")

                print("="*50)
                print(f"[Laissez] SAVE THIS FOR FUTURE USE")
                print(f"[Laissez] Wallet address: {wallet.address}")
                print(f"[Laissez] Wallet private key: {wallet.key.hex()}")
                print("="*50)
            else:
                raise ValueError("No wallet found and user did not want to create one")
    return wallet


"""
==========
LAISSEZ
==========

EXAMPLE: Paid MCP Server with Laissez

This example shows how to create a paid MCP server with Laissez.

Pre-requisites:
* Update `WALLET_PRIVATE_KEY` in `.env` equal to the private key for your EVM-compatible account in which you will receive payments
* Ensure that the names of the paid tools in the `paid_tools` list match the names of the tools in the MCP server

To run:
* Run this file from the root of the project by running `uv run examples/mcp/paid-mcp-server.py`.
"""

from fastmcp import FastMCP
import random
from laissez.server import create_paid_mcp_app, PaidTool
from dotenv import load_dotenv

load_dotenv()

mcp = FastMCP("Dice Roller")

@mcp.tool
def roll_die() -> int:
    return random.randint(1, 6)

@mcp.tool
def multiply(a: int, b: int) -> int:
    return a * b


paid_tools = [
    PaidTool(name="roll_die", price=0.001, description="Roll a die and return the result"),
    PaidTool(name="multiply", price=0.005, description="Multiply two numbers and return the result"),
]


if __name__ == "__main__":
    import uvicorn
    import asyncio 

    app = asyncio.run(create_paid_mcp_app(mcp, paid_tools))
    uvicorn.run(app, host="127.0.0.1", port=8000)


from fastapi import FastAPI
from x402.fastapi.middleware import require_payment
import random

app = FastAPI()

app.middleware("http")(
    require_payment(
        path="/roll-dice",
        price="$0.001",
        pay_to_address="0x58635e118B55F67f75B4dFD44f3cd612eC536e0a",
        network="base-sepolia",
        description="Roll a dice"
    )
)

app.middleware("http")(
    require_payment(
        path="/multiply",
        price="$0.005",
        pay_to_address="0x58635e118B55F67f75B4dFD44f3cd612eC536e0a",
        network="base-sepolia",
        description="Multiply two numbers"
    )
)

@app.get("/roll-dice")
def roll_dice():
    dice_roll = random.randint(1, 6)
    return { "dice_roll": dice_roll } 


@app.get("/multiply")
def multiply(a: int, b: int):
    return { "result": a * b }